#!/usr/bin/env python3
"""This module tests the info module"""
import unittest
import yaml
import libpkpass.commands.cli as cli
import libpkpass.commands.info as info
from libpkpass.errors import CliArgumentError
from .basetest.basetest import captured_output, patch_args, ERROR_MSGS

class InfoTests(unittest.TestCase):
    """This class tests the info class"""
    def setUp(self):
        self.check_dict = {
            'Metadata': {
                'Authorizer': 'y',
                'Creator': 'r1',
                'Description': 'y',
                'Name': 'test',
                'Schemaversion': 'v2',
                'Signature': 'None'
            },
            'Recipients': 'r1',
            'Total Recipients': 1,
        }

    def test_info(self):
        """Test what info shows on a password"""
        with patch_args(subparser_name='info', identity='r1',
                        nopassphrase="true", pwname='test'):
            with captured_output() as (out, _):
                info.Info(cli.Cli())
        output = yaml.safe_load(out.getvalue())
        del output['Earliest distribute timestamp']
        self.assertDictEqual(output, self.check_dict)

    def test_info_no_pass(self):
        """Test what happens when pwname is not supplied"""
        with patch_args(subparser_name='info', identity='r1',
                        nopassphrase="true", pwname=None):
            with self.assertRaises(CliArgumentError) as context:
                info.Info(cli.Cli())
        self.assertEqual(context.exception.msg, ERROR_MSGS['pwname'])

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""This Module handles the crypto functions i.e. encryption and decryption"""
from base64 import urlsafe_b64decode, urlsafe_b64encode
from tempfile import NamedTemporaryFile
from os import unlink
from hashlib import sha256
from shutil import get_terminal_size
from subprocess import Popen, PIPE, STDOUT, DEVNULL
from pem import parse_file
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from libpkpass.util import color_prepare
from libpkpass.errors import EncryptionError, DecryptionError, SignatureCreationError, X509CertificateError

    ##############################################################################
def handle_python_strings(string):
    """handles py2/3 incompatiblities"""
    ##############################################################################
    return string if isinstance(string, bytes) else string.encode("UTF-8")

    ##############################################################################
def pk_encrypt_string(plaintext_string, certificate):
    """ Encrypt and return a base 64 encoded string for the provided identity"""
    ##############################################################################
    plaintext_derived_key = Fernet.generate_key()
    if isinstance(certificate, bytes):
        with NamedTemporaryFile(delete=False) as fname:
            fname.write(certificate)
            cert_file_path = fname.name
    else:
        cert_file_path = certificate['certificate_path']
    command = ['openssl', 'rsautl', '-inkey', cert_file_path, '-certin', '-encrypt', '-pkcs']
    proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, _ = proc.communicate(input=plaintext_derived_key)
    if isinstance(certificate, bytes):
        unlink(fname.name)

    if proc.returncode != 0:
        raise EncryptionError("Error encrypting derived key: %s" % stdout)
    ciphertext_derived_key = stdout

    encrypted_string = Fernet(plaintext_derived_key).encrypt(
        handle_python_strings(plaintext_string)
    )
    return (encrypted_string, urlsafe_b64encode(handle_python_strings(ciphertext_derived_key)))

    ##############################################################################
def get_card_info():
    ##############################################################################
    command = ['pkcs11-tool', '-L']
    proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, _ = proc.communicate()
    return (handle_python_strings(stdout).split(b'Slot'), stdout)

    ##############################################################################
def print_card_info(card_slot, identity, verbosity, color, theme_map):
    """Inform the user what card is selected"""
    ##############################################################################
    if 'key_path' not in identity:
        out_list, stdout = get_card_info()
        if verbosity > 1:
            print_all_slots(stdout, color, theme_map)
        for out in out_list[1:]:
            stripped = out.decode("UTF-8").strip()
            if int(stripped[0]) == int(card_slot):
                verbosity = verbosity + 1 if verbosity < 2 else 2
                stripped = "Using Slot" + ("\n").join(stripped.split('\n')[:verbosity])
                print("%s" % (color_prepare(stripped, "info", color, theme_map)))

    ##############################################################################
def print_all_slots(slot_info, color, theme_map):
    """Print all slots and cards available"""
    ##############################################################################
    columns = int(get_terminal_size().columns) // 4
    print(color_prepare("#" * columns, "debug", color, theme_map),
          handle_python_strings(slot_info).decode("UTF-8").strip(),
          color_prepare("#" * columns, "debug", color, theme_map), sep="\n")

    ##############################################################################
def pk_decrypt_string(ciphertext_string, ciphertext_derived_key, identity, passphrase, card_slot=None):
    """ Decrypt a base64 encoded string for the provided identity"""
    ##############################################################################
    ciphertext_derived_key = handle_python_strings(ciphertext_derived_key)
    if 'key_path' in identity:
        command = ['openssl', 'rsautl', '-inkey', identity['key_path'], '-decrypt', '-pkcs']
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate(
            input=urlsafe_b64decode(ciphertext_derived_key))
        plaintext_derived_key = stdout
    else:
        # We've got to use pkcs15-crypt for PIV cards, and it only supports pin on
        # command line (insecure) or via stdin.  So, we have to put ciphertext into
        # a file for pkcs15-crypt to read.  YUCK!
        with NamedTemporaryFile(delete=False) as fname:
            fname.write(urlsafe_b64decode(ciphertext_derived_key))
        command = ['pkcs15-crypt', '--decipher', '--raw', '--pkcs', '--input', fname.name]
        if card_slot is not None:
            command.extend(['--reader', str(card_slot)])
        command.extend(['--pin', '-'])
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=DEVNULL)
        stdout, _ = proc.communicate(input=passphrase.encode('UTF-8'))
        unlink(fname.name)
        try:
            plaintext_derived_key = stdout
        except IndexError as err:
            raise DecryptionError(stdout) from err

    if proc.returncode != 0:
        raise DecryptionError(stdout)

    return Fernet(plaintext_derived_key).decrypt(
        handle_python_strings(ciphertext_string)
    ).decode("UTF-8")

    ##############################################################################
def pk_sign_string(string, identity, passphrase, card_slot=None):
    """ Compute the hash of string and create a digital signature """
    ##############################################################################
    stringhash = sha256(string.encode("UTF-8")).hexdigest()
    if 'key_path' in identity:
        command = ['openssl', 'rsautl', '-sign', '-inkey', identity['key_path']]
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate(input=stringhash.encode('UTF-8'))
        signature = urlsafe_b64encode(handle_python_strings(stdout))
    else:
        # We've got to use pkcs15-crypt for PIV cards, and it only supports pin on
        # command line (insecure) or via stdin.  So, we have to put signature text
        # into a file for pkcs15-crypt to read.  YUCK!
        with NamedTemporaryFile(delete=False) as fname:
            fname.write(stringhash.encode('UTF-8'))
        out = NamedTemporaryFile(delete=False)
        command = ['pkcs15-crypt', '--sign', '-i', fname.name, '-o', out.name, '--pkcs1', '--raw']
        if card_slot is not None:
            command.extend(['--reader', str(card_slot)])
        command.extend(['--pin', '-'])
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate(input=passphrase.encode('UTF-8'))
        out.close()

        with open(out.name, 'rb') as sigfile:
            signature = urlsafe_b64encode(handle_python_strings(sigfile.read()))

        unlink(fname.name)
        unlink(out.name)

    if proc.returncode != 0:
        raise SignatureCreationError(stdout)

    return signature

    ##############################################################################
def pk_verify_signature(string, signature, identity):
    """ Compute the hash of string and verify the digital signature """
    ##############################################################################
    stringhash = sha256(string.encode("UTF-8")).hexdigest()
    for cert in parse_file(identity['certificate_path']):
        with NamedTemporaryFile(delete=False) as fname:
            fname.write(cert.as_bytes())
        command = ['openssl', 'rsautl', '-inkey', fname.name, '-certin', '-verify']
        proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        stdout, _ = proc.communicate(input=urlsafe_b64decode(handle_python_strings(signature)))
        unlink(fname.name)
        if stdout.decode("UTF-8") == stringhash:
            return True

    return stdout.decode("UTF-8") == stringhash

    ##############################################################################
def pk_verify_chain(certificate, cabundle):
    """ Verify the publickey trust chain against a CA Bundle and return True if valid"""
    ##############################################################################
    command = ['openssl', 'verify', '-CAfile', cabundle]
    proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, _ = proc.communicate(input=certificate)

    return stdout.decode("UTF-8").rstrip() == "stdin: OK"

    ##############################################################################
def get_cert_fingerprint(cert):
    """ Return the modulus of the x509 certificate of the identity """
    ##############################################################################
    # SHA1 Fingerprint=F9:9D:71:54:55:BE:99:24:6A:5E:E0:BB:48:F9:63:AE:A2:05:54:98
    return get_cert_element(cert, 'fingerprint').split('=')[1]

    ##############################################################################
def get_cert_subject(cert):
    """ Return the subject DN of the x509 certificate of the identity """
    ##############################################################################
    # subject= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return ' '.join(get_cert_element(cert, 'subject').split(' ')[1:])

    ##############################################################################
def get_cert_issuer(cert):
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    # issuer= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return ' '.join(get_cert_element(cert, 'issuer').split(' ')[1:])

    ##############################################################################
def get_cert_enddate(cert):
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_cert_element(cert, 'enddate').split('=')[1]

    ##############################################################################
def get_cert_issuerhash(cert):
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_cert_element(cert, 'issuer_hash')

    ##############################################################################
def get_cert_subjecthash(cert):
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_cert_element(cert, 'subject_hash')

    ##############################################################################
def get_cert_element(cert, element):
    """ Return an arbitrary element of an x509 certificate """
    ##############################################################################
    command = ['openssl', 'x509', '-noout', ("-%s" % element)]
    proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, _ = proc.communicate(input=cert)
    if proc.returncode != 0:
        raise X509CertificateError(stdout)

    try:
        return stdout.decode("UTF-8").rstrip()
    except IndexError as err:
        raise X509CertificateError(stdout) from err

    ##############################################################################
def get_card_element(element):
    """Return an arbitrary element of a pcks15 capable device"""
    ##############################################################################
    command = ['pkcs15-tool', '--read-certificate', '1']
    proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, _ = proc.communicate()
    if stdout.decode('utf-8').strip().lower() == 'no smart card readers found.':
        raise X509CertificateError('Smartcard not detected')
    return get_cert_element(stdout, element)

    ##############################################################################
def get_card_fingerprint():
    """ Return the modulus of the x509 certificate of the identity """
    ##############################################################################
    # SHA1 Fingerprint=F9:9D:71:54:55:BE:99:24:6A:5E:E0:BB:48:F9:63:AE:A2:05:54:98
    return get_card_element('fingerprint').split('=')[1]

    ##############################################################################
def get_card_subject():
    """ Return the subject DN of the x509 certificate of the identity """
    ##############################################################################
    # subject= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return ' '.join(get_card_element('subject').split(' ')[1:])

    ##############################################################################
def get_card_issuer():
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    # issuer= /C=US/O=Entrust/OU=Certification Authorities/OU=Entrust Managed Services SSP CA
    return ' '.join(get_card_element('issuer').split(' ')[1:])

    ##############################################################################
def get_card_startdate():
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_card_element('startdate').split('=')[1]

    ##############################################################################
def get_card_enddate():
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_card_element('enddate').split('=')[1]

    ##############################################################################
def get_card_issuerhash():
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_card_element('issuer_hash')

    ##############################################################################
def get_card_subjecthash():
    """ Return the issuer DN of the x509 certificate of the identity """
    ##############################################################################
    return get_card_element('subject_hash')

    ##############################################################################
def sk_encrypt_string(plaintext_string, key):
    """ Symmetrically Encrypt and return a base 64 encoded string using the provided secret"""
    ##############################################################################
    fern = Fernet(hash_password(key))
    encrypted_string = fern.encrypt(plaintext_string.encode('UTF-8'))
    return urlsafe_b64encode(handle_python_strings(encrypted_string))

    ##############################################################################
def sk_decrypt_string(ciphertext_string, key):
    """ Symmetrically Decrypt a base64 encoded string using the provided key"""
    ##############################################################################
    fern = Fernet(hash_password(key))
    ciphertext_string = handle_python_strings(ciphertext_string)
    try:
        plaintext_string = fern.decrypt(urlsafe_b64decode(handle_python_strings(ciphertext_string)))
        return plaintext_string.decode("UTF-8")
    except InvalidToken:
        # handle some legacy stuff from NCCS
        try:
            plaintext_string = fern.decrypt(ciphertext_string)
            return plaintext_string.decode("UTF-8")
        except InvalidToken as err:
            raise DecryptionError("Incorrect Password") from err

    ##############################################################################
def hash_password(password):
    """Hash people's password"""
    ##############################################################################
    password = handle_python_strings("%s" % password)
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=b"salt",
                     iterations=100000, backend=default_backend())
    return urlsafe_b64encode(handle_python_strings(kdf.derive(password)))

    ##############################################################################
def puppet_password(eyaml, password):
    """Give eyaml password"""
    ##############################################################################
    command = [eyaml, 'encrypt', '-s', password, '-o', 'string']
    proc = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
    stdout, _ = proc.communicate()
    return stdout.decode().strip()



#!/usr/bin/env python3
"""This Module is the base module for custom errors to allow for reduced tracebacks and
more actionable items on the user's side"""

class PKPassError(Exception):
    """Base class for pkpass exceptions."""

    def __init__(self, arg):
        self.msg = arg

class BlankPasswordError(PKPassError):
    def __init__(self):
        self.msg = "User Provided password is blank or only spaces"

class CliArgumentError(PKPassError):
    pass

class ConfigParseError(PKPassError):
    pass

class DecryptionError(PKPassError):
    pass

class EncryptionError(PKPassError):
    pass

class EscrowError(PKPassError):
    def __init__(self, field, constant, value):
        self.msg = "%s, must be greater than %s, current value: %s" % (
            field, constant, value)

class FileOpenError(PKPassError):
    def __init__(self, value, reason):
        self.msg = "File %s found in config, could not be opened due to %s" % (
            value, reason)

class GroupDefinitionError(PKPassError):
    def __init__(self, value):
        self.msg = "Group %s is not defined in the config" % (value)

class JsonArgumentError(PKPassError):
    def __init__(self, value, reason):
        self.msg = "Parse error for '%s' because '%s'" %  (value, reason)

class LegacyImportFormatError(PKPassError):
    def __init__(self):
        self.msg = "Passwords in import file not in key:value notation"

class NotARecipientError(PKPassError):
    pass

class NotThePasswordOwnerError(PKPassError):
    def __init__(self, identity, owner, pwname):
        self.msg = "User '%s' is not the owner of password '%s', not overwriting;\n\t\
please use another password name or ask the owner (%s) to distribute to you" % (identity, pwname, owner)

class NullRecipientError(PKPassError):
    def __init__(self):
        self.msg = "There is a blank Recipient in the list, please check for trailing commas"

class PasswordIOError(PKPassError):
    pass

class PasswordMismatchError(PKPassError):
    def __init__(self):
        self.msg = "Passwords do not match"

class PasswordValidationError(PKPassError):
    def __init__(self, field, value):
        self.msg = "Error validating password field: %s with value %s" % (
            field, value)

class RulesMapError(PKPassError):
    def __init__(self, reason):
        self.msg = reason

class RSAKeyError(PKPassError):
    pass

class SignatureCreationError(PKPassError):
    pass

class TrustChainVerificationError(PKPassError):
    pass

class X509CertificateError(PKPassError):
    pass

class YamlFormatError(PKPassError):
    def __init__(self, value, reason):
        self.msg = "Error%s due to: %s" % (value, reason)



"""This Modules allows for distributing created passwords to other users"""
from os import path
from tqdm import tqdm
import libpkpass.util as util
from libpkpass.commands.command import Command
from libpkpass.passworddb import PasswordDB
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError

    ####################################################################
class Distribute(Command):
    """This Class implements the CLI functionality for ditribution"""
    ####################################################################
    name = 'distribute'
    description = 'Distribute existing password entry/ies to another entity [matching uses python fnmatch]'
    selected_args = Command.selected_args + ['pwname', 'pwstore', 'users', 'groups', 'stdin',
                                             'min_escrow', 'escrow_users', 'keypath', 'nopassphrase',
                                             'nosign', 'card_slot', 'noescrow']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        passworddb = PasswordDB()
        passworddb.load_from_directory(self.args['pwstore'])
        filtered_pdb = util.dictionary_filter(
            path.join(self.args['pwstore'], self.args['pwname']),
            passworddb.pwdb,
            [self.args['identity'], 'recipients']
        )
        self.recipient_list.append(str(self.args['identity']))
        self.recipient_list = list(set(self.recipient_list))
        print("The following users will receive the password:")
        print(", ".join(self.recipient_list))
        print("The following password files have matched:")
        print(*filtered_pdb.keys(), sep="\n")
        correct_distribution = input("Are these lists correct? (y/N) ")
        if correct_distribution and correct_distribution.lower()[0] == 'y':
            passworddb.pwdb = filtered_pdb
            for dist_pass, _ in tqdm(passworddb.pwdb.items()):
                password = PasswordEntry()
                password.read_password_data(dist_pass)
                if self.args['identity'] in password.recipients.keys():
                    # we shouldn't modify escrow on distribute
                    self.args['min_escrow'] = None
                    self.args['escrow_users'] = None
                    plaintext_pw = password.decrypt_entry(
                        self.identities.iddb[self.args['identity']],
                        passphrase=self.passphrase,
                        card_slot=self.args['card_slot'])

                    password.read_password_data(dist_pass)
                    password.add_recipients(secret=plaintext_pw,
                                            distributor=self.args['identity'],
                                            recipients=self.recipient_list,
                                            identitydb=self.identities,
                                            passphrase=self.passphrase,
                                            card_slot=self.args['card_slot'],
                                            pwstore=self.args['pwstore']
                                           )

                    password.write_password_data(dist_pass)
        else:
            print("Exiting due to wrong password list")

        ####################################################################
    def _validate_args(self):
        ####################################################################
        for argument in ['pwname', 'keypath']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(
                    "'%s' is a required argument" % argument)



"""This Module allows for the export of passwords for the purpose of importing to a new card"""
from os import path, walk
import getpass
from tqdm import tqdm
from libpkpass.commands.command import Command
from libpkpass.passworddb import PasswordDB
from libpkpass.errors import CliArgumentError, PasswordMismatchError

    ####################################################################
class Export(Command):
    """This Class implements the cli functionality of export"""
    ####################################################################
    name = 'export'
    description = 'Export passwords that you have access to and encrypt with aes'
    selected_args = Command.selected_args + ['pwfile', 'stdin', 'nopassphrase',
                                             'card_slot', 'nocrypto']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        myidentity = self.identities.iddb[self.args['identity']]
        passworddb = PasswordDB()
        crypt_pass = False
        for fpath, _, files in walk(self.args['pwstore']):
            for passwordname in files:
                passwordpath = path.join(fpath, passwordname)
                passworddb.load_password_data(passwordpath)
        if not self.args['nocrypto']:
            crypt_pass = getpass.getpass("Please enter a password for the encryption: ")
            verify_pass = getpass.getpass("Please enter again for verification: ")
            if crypt_pass != verify_pass:
                raise PasswordMismatchError()

        self._iterate_pdb(myidentity, passworddb, crypt_pass)

        ####################################################################
    def _iterate_pdb(self, myidentity, passworddb, crypt_pass=False):
        """ Iterate through the passwords that we can decrypt """
        ####################################################################
        uid = myidentity['uid']
        all_passwords = {k:v for (k, v) in passworddb.pwdb.items() if uid in v.recipients.keys()}
        for _, password in tqdm(all_passwords.items()):
            plaintext_pw = password.decrypt_entry(
                identity=myidentity,
                passphrase=self.passphrase,
                card_slot=self.args["card_slot"])
            password.recipients[uid]['encrypted_secret'] = plaintext_pw.encode("UTF-8")
            password.write_password_data(self.args['pwfile'],
                                         overwrite=False,
                                         encrypted_export=not self.args['nocrypto'],
                                         password=crypt_pass,
                                         export=True)

        ####################################################################
    def _validate_args(self):
        """ Ensure arguments are appropriate for this command           """
        ####################################################################
        for argument in ['pwfile', 'keypath']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(
                    "'%s' is a required argument" % argument)

        ####################################################################
    def _validate_combinatorial_args(self):
        ####################################################################
        pass



#!/usr/bin/env python3
"""This Module tests the passworddb module"""
import unittest
from libpkpass.passworddb import PasswordDB

class TestBasicFunction(unittest.TestCase):
    """This class tests the passworddb class"""
    def setUp(self):
        self.file1 = 'test/passwords/testpassword'
        self.file2 = 'test/scratch/testpassword'

    def test_read_write(self):
        """Test read write to the password db"""
        passworddb = PasswordDB()
        passwordentry = passworddb.load_password_data(self.file1)
        passworddb.pwdb[self.file2] = passworddb.pwdb[self.file1]
        passworddb.save_password_data(self.file2, overwrite=True)

        with open(self.file1, 'r') as file1:
            with open(self.file2, 'r') as file2:
                self.assertTrue(file1.read() == file2.read())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""This module tests the distribute module"""
import builtins
import unittest
import mock
import libpkpass.commands.cli as cli
import libpkpass.commands.distribute as distribute
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS

class DistributeTests(unittest.TestCase):
    """This class tests the distribute class"""
    def test_recipient_not_in_database(self):
        """Test what happens with a recipient is not within the appropriate directory"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='distribute', identity='bleh', nopassphrase='true',
                            pwname='test'):
                distribute.Distribute(cli.Cli())
        self.assertEqual(
            context.exception.msg,
            ERROR_MSGS['rep']
        )

    def test_distribute_cli_error(self):
        """Test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='distribute', identity='r1', nopassphrase='true',
                            pwname=None):
                distribute.Distribute(cli.Cli())
        self.assertEqual(context.exception.msg, ERROR_MSGS['pwname'])

    def test_distribute_success(self):
        """Test a successful distribute"""
        ret = True
        try:
            with patch_args(subparser_name='distribute', identity='r1', nopassphrase='true',
                            pwname='test'):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    distribute.Distribute(cli.Cli())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)


if __name__ == '__main__':
    unittest.main()



"""This Module defines what a password contains"""
from uuid import uuid1
from time import time
from datetime import datetime
from os import path, makedirs
from dateutil import parser
import yaml
from tqdm import tqdm
from libpkpass.escrow import pk_split_secret
from libpkpass.errors import NotARecipientError, DecryptionError, PasswordIOError, YamlFormatError,\
    X509CertificateError
import libpkpass.crypto as crypto

    #######################################################################
class PasswordEntry():
    """ Password entry object.  Contains information about a password and
    related metadata, as well as a list of individually encrypted strings"""
    ##########################################################################
    def __init__(self, **kwargs):

        self.metadata = {'name': None,
                         'description': None,
                         'creator': None,
                         'authorizer': None,
                         'signature': None,
                         'schemaVersion': 'v2'}

        self.metadata.update(kwargs)
        self.recipients = {}
        self.escrow = {}

        ####################################################################
    def __getitem__(self, key):
        """Grab items like with dictionaries"""
        ####################################################################
        return getattr(self, key)

        ####################################################################
    def read_escrow(self, filename):
        """Read values from escrow section"""
        ####################################################################
        try:
            with open(filename, 'r') as fname:
                password_data = yaml.safe_load(fname)
                if 'escrow' in password_data.keys():
                    return password_data['escrow']
                return {}
        except (OSError, IOError, yaml.scanner.ScannerError, yaml.parser.ParserError):
            #we actually don't care here.. file might not exist yet
            return {}

        #######################################################################
    def process_escrow_map(
            self,
            escrow_map,
            split_secret=None,
            distributor=None,
            recipients=None,
            identitydb=None,
            encryption_algorithm='rsautl',
            passphrase=None,
            card_slot=None,
            escrow_users=None,
            minimum=None):
        """Process the escrow user map into escrow users"""
        #######################################################################
        escrow_guid = str(uuid1()).replace('-', '')
        if escrow_map:
            for key, value in escrow_map.items():
                if set(value['recipients']) == set(escrow_users):
                    escrow_guid = key
        self.escrow[escrow_guid] = {
            'metadata': {
                'minimum_escrow': minimum,
                'creator': distributor
                },
            'recipients':{}
            }
        i = 0
        for escrow_user in escrow_users:
            if escrow_user not in recipients:
                self.escrow[escrow_guid]['recipients'][escrow_user] = self._add_recipient(
                    escrow_user, split_secret[i], distributor,
                    identitydb, encryption_algorithm, passphrase, card_slot)
                i += 1


        #######################################################################
    def add_escrow(
            self,
            secret=None,
            escrow_users=None,
            minimum=None,
            pwstore=None):
        """ Add escrow users to the recipient list of this password object"""
        #######################################################################
        split_secret = pk_split_secret(secret, escrow_users, minimum)
        return (self.read_escrow(path.join(pwstore, self.metadata['name'])), split_secret)

        #######################################################################
    def add_recipients(
            self,
            secret=None,
            distributor=None,
            recipients=None,
            identitydb=None,
            encryption_algorithm='rsautl',
            passphrase=None,
            card_slot=None,
            escrow_users=None,
            minimum=None,
            pwstore=None):
        """ Add recipients to the recipient list of this password object"""
        #######################################################################
        if recipients is None:
            recipients = []

        new_recipients = {r:self._add_recipient(
            r, secret, distributor, identitydb, encryption_algorithm, passphrase, card_slot
        ) for r in tqdm(recipients, leave=False)}
        self.recipients.update(new_recipients)
        if escrow_users:
            #escrow_users may now be none after the set operations
            try:
                if (len(escrow_users) > 3) and (len(list((set(escrow_users) - set(recipients)))) < 3):
                    print("warning: min_escrow requirement not met after removing password recipients from escrow user list")
                    return
                escrow_map, split_secret = self.add_escrow(
                    secret=secret,
                    escrow_users=escrow_users,
                    minimum=minimum,
                    pwstore=pwstore)
                self.process_escrow_map(
                    escrow_map,
                    split_secret=split_secret,
                    distributor=distributor,
                    recipients=recipients,
                    identitydb=identitydb,
                    encryption_algorithm=encryption_algorithm,
                    passphrase=passphrase,
                    card_slot=card_slot,
                    escrow_users=escrow_users,
                    minimum=minimum)
            except ValueError as err:
                print("Warning cannot create escrow shares, reason: %s" % err)
                print("Your password has been created without escrow capabilities")


        #######################################################################
    def _add_recipient(
            self,
            recipient,
            secret=None,
            distributor=None,
            identitydb=None,
            encryption_algorithm='rsautl',
            passphrase=None,
            card_slot=None,
    ):
        """Add recipient or sharer to list"""
        #######################################################################
        try:
            encrypted_secrets = {}
            for cert in identitydb.iddb[recipient]['certs']:
                if encryption_algorithm == 'rsautl':
                    if 'key_path' in identitydb.iddb[recipient].keys():
                        (encrypted_secret, encrypted_derived_key) = crypto.pk_encrypt_string(
                            secret, identitydb.iddb[recipient])
                    else:
                        (encrypted_secret, encrypted_derived_key) = crypto.pk_encrypt_string(
                            secret, cert['cert_bytes'])
                encrypted_secrets[cert['fingerprint']] = {
                    'encrypted_secret': encrypted_secret,
                    'derived_key': encrypted_derived_key,
                    'recipient_hash': cert['subjecthash'],
                }
            try:
                distributor_hash = crypto.get_card_subjecthash()
            except X509CertificateError:
                distributor_hash = identitydb.iddb[distributor]['certs'][0]['subjecthash']
            recipient_entry = {
                'encrypted_secrets': encrypted_secrets,
                'encryption_algorithm': encryption_algorithm,
                'timestamp': time(),
                'distributor': distributor,
                'distributor_hash': distributor_hash,
            }
            message = self._create_signable_string(recipient_entry)
            recipient_entry['signature'] = crypto.pk_sign_string(
                message,
                identitydb.iddb[distributor],
                passphrase, card_slot
            )

            return recipient_entry
        except KeyError as err:
            raise NotARecipientError(
                "Identity '%s' is not on the recipient list for password '%s'" %
                (recipient, self.metadata['name'])) from err

        #######################################################################
    def decrypt_entry(self, identity=None, passphrase=None, card_slot=None):
        """ Decrypt this password entry for a particular identity
        (usually the user) """
        #######################################################################
        try:
            recipient_entry = self.recipients[identity['uid']]
        except KeyError as err:
            raise NotARecipientError(
                "Identity '%s' is not on the recipient list for password '%s'" %
                (identity['uid'], self.metadata['name'])) from err

        try:
            # support old yaml format
            return crypto.pk_decrypt_string(
                recipient_entry['encrypted_secret'],
                recipient_entry['derived_key'],
                identity,
                passphrase,
                card_slot)
        except KeyError:
            try:
                # support rsa
                if 'key_path' in identity.keys():
                    for _, value in recipient_entry['encrypted_secrets'].items():
                        return crypto.pk_decrypt_string(
                            value['encrypted_secret'],
                            value['derived_key'],
                            identity,
                            passphrase
                        )
                else:
                    cert_key = crypto.get_card_fingerprint()
                    return crypto.pk_decrypt_string(
                        recipient_entry['encrypted_secrets'][cert_key]['encrypted_secret'],
                        recipient_entry['encrypted_secrets'][cert_key]['derived_key'],
                        identity,
                        passphrase,
                        card_slot)
            except DecryptionError as err:
                msg = create_error_message(recipient_entry['timestamp'], card_slot)
                raise DecryptionError(
                    "Error decrypting password named '%s'. %s" %
                    (self.metadata['name'], msg)) from err
            except KeyError as err:
                raise DecryptionError(
                    "Error decrypting password named '%s'. Appropriate private key not found" %
                    self.metadata['name']) from err
        except DecryptionError as err:
            msg = create_error_message(recipient_entry['timestamp'], card_slot)
            raise DecryptionError("Error decrypting password named '%s'. %s" %
                                  (self.metadata['name'], msg)) from err

        #######################################################################
    def verify_entry(self, uid=None, iddb=None):
        """ Check to see if all signatures and user certificates are okay for
        the user accessing this password entry """
        #######################################################################
        identitydb = iddb.iddb
        recipient_entry = self.recipients[uid].copy()
        distributor = recipient_entry['distributor']
        iddb.verify_identity(distributor, [])
        signature = recipient_entry.pop('signature')
        message = self._create_signable_string(recipient_entry)
        sig_ok = crypto.pk_verify_signature(
            message, signature, identitydb[distributor])
        return {
            'distributor': distributor,
            'sigOK': sig_ok,
            'certOK': identitydb[distributor]['certs'][0]['verified']
        }

        #######################################################################
    def _create_signable_string(self, recipient_entry):
        """ Helper function to create a string out of the information
        this PasswordEntry contains"""
        #######################################################################
        message = ""
        for key in sorted(recipient_entry.keys()):
            if key == 'timestamp':
                recipient_entry[key] = "{0:12.2f}".format(float(recipient_entry[key]))
            if key not in ['signature', 'encrypted_secrets']:
                if isinstance(recipient_entry[key], bytes):
                    recipient_entry[key] = recipient_entry[key].decode('UTF-8')
                message = message + str(key) + str(recipient_entry[key])
        return message

        #######################################################################
    def todict(self):
        """Returns a dictionary"""
        #######################################################################
        return vars(self)

        #######################################################################
    def validate(self):
        """This function validates a password object"""
        #######################################################################
        # for field in ['field1', 'field2']:
        #  if weird: raise PasswordValidationError(field, value)
        return True

        #######################################################################
    def __repr__(self):
        #######################################################################
        return "%s(%r)" % (self.__class__, self.__dict__)

        #######################################################################
    def __str__(self):
        #######################################################################
        return "%r" % self.__dict__

        ##########################################################################
    def read_password_data(self, filename):
        """ Open a password file, load passwords and read metadata"""
        ##########################################################################
        try:
            with open(filename, 'r') as fname:
                password_data = yaml.safe_load(fname)
                self.metadata = password_data['metadata']
                self.recipients = password_data['recipients']
                if 'escrow' in password_data:
                    self.escrow = password_data['escrow']
            self.validate()
        except (OSError, IOError) as error:
            raise PasswordIOError(
                "Error Opening %s due to %s" %
                (filename, error.strerror)) from error
        except (yaml.scanner.ScannerError, yaml.parser.ParserError) as error:
            raise YamlFormatError(str(error.problem_mark), error.problem) from error

        ############################################################################################
    def write_password_data(self, filename, overwrite=False,
                            export=False,
                            encrypted_export=False,
                            password=None):
        """Write password data to a password file """
        ############################################################################################
        self.validate()
        # check if the file is in the base password directory
        incwd = path.basename(filename) == filename
        # if not ensure path exists
        try:
            if not incwd and not path.isdir(path.dirname(filename)):
                makedirs(path.dirname(filename))
            passdata = {key: value for key, value in self.todict().items() if value}
            open_method = 'a' if export else 'w'
            with open(filename, open_method) as fname:
                if encrypted_export:
                    encrypted = crypto.sk_encrypt_string(
                        yaml.safe_dump(passdata, default_flow_style=False),
                        password)
                    fname.write(encrypted.decode() + "\n")
                else:
                    fname.write(yaml.safe_dump(passdata, default_flow_style=False))
        except (OSError, IOError) as error:
            raise PasswordIOError("Error creating '%s'" % filename) from error

def create_error_message(recipient_timestamp, card_slot):
    card_start = crypto.get_card_startdate()
    card_start = datetime.timestamp(parser.parse(card_start))
    distribute_time = float(recipient_timestamp)
    # Slots are indexed at 0 so when enumerating you add 1
    # There is also an additional information line so add 1 again
    if int(card_slot) + 2 > len(crypto.get_card_info()[0]):
        msg = 'Attempting to use card slot that is not connected'
    elif distribute_time < card_start:
        msg = 'Password distributed before this certificate was created'
    else:
        msg = "Perhaps a bad pin/passphrase?"
    return msg



#!/usr/bin/env python3
"""This module tests the update module"""
import getpass
import builtins
import unittest
import mock
import libpkpass.commands.cli as cli
import libpkpass.commands.update as update
from libpkpass.errors import DecryptionError
from .basetest.basetest import patch_args

class UpdateTests(unittest.TestCase):
    """This class tests the update class"""
    def test_update_success(self):
        """Test a successful update of a password"""
        ret = True
        try:
            with patch_args(subparser_name='update', identity='r3', nopassphrase='true',
                            pwname='gentest'):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    with mock.patch.object(getpass, 'getpass', lambda _: 'y'):
                        update.Update(cli.Cli())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)



#!/usr/bin/env python3
"""This is a super class to handle commonalities between the cli and interpreter"""
from argparse import ArgumentParser
from os import path
import libpkpass.util as util
import libpkpass.commands.card as card
import libpkpass.commands.clip as clip
import libpkpass.commands.create as create
import libpkpass.commands.delete as delete
import libpkpass.commands.distribute as distribute
import libpkpass.commands.export as export
import libpkpass.commands.fileimport as pkimport
import libpkpass.commands.generate as generate
import libpkpass.commands.info as info
import libpkpass.commands.list as pklist
import libpkpass.commands.listrecipients as listrecipients
import libpkpass.commands.modify as modify
import libpkpass.commands.recover as recover
import libpkpass.commands.rename as rename
import libpkpass.commands.show as show
import libpkpass.commands.populate as populate
import libpkpass.commands.update as update
import libpkpass.commands.verifyinstall as verifyinstall

    ##############################################################################
class PkInterface():
    """ Class for parsing command line.  Observes subclasses of Command to Register
    those commands in the actions list.                                        """
    ##############################################################################
        ####################################################################
    def __init__(self):
        """ Intialization function for class. Register all subcommands   """
        ####################################################################
        # Hash of registered subparser actions, mapping string to actual subparser
        self.actions = {}
        home = path.expanduser("~")
        defaultrc = None
        for file in ['.pkpassrc', '.pkpassrc.yml', '.pkpassrc.yaml']:
            if path.isfile(path.join(home, file)):
                defaultrc = path.join(home, file)
        if not defaultrc:
            defaultrc = path.join(home, '.pkpassrc')
        self.parser = ArgumentParser(
            description='Public Key Password Manager')
        self.parser.set_default_subparser = util.set_default_subparser
        self.parser.add_argument(
            '--config', type=str,
            help="Path to a PKPass configuration file.  Defaults to '~/.pkpassrc{,.yml,.yaml}'",
            default=defaultrc
        )
        self.parser.add_argument('--debug', action='store_true',
                                 help="Errors are more verbose")
        self.subparsers = self.parser.add_subparsers(
            help='sub-commands', dest='subparser_name')

        card.Card(self)
        clip.Clip(self)
        create.Create(self)
        delete.Delete(self)
        distribute.Distribute(self)
        export.Export(self)
        generate.Generate(self)
        pkimport.Import(self)
        info.Info(self)
        pklist.List(self)
        listrecipients.Listrecipients(self)
        modify.Modify(self)
        recover.Recover(self)
        rename.Rename(self)
        show.Show(self)
        populate.Populate(self)
        update.Update(self)
        verifyinstall.VerifyInstall(self)

        ####################################################################
    def register(self, command_obj, command_name, command_description):
        """ Register command objects and names using an observer pattern """
        ####################################################################
        self.actions[command_name] = command_obj
        parser = self.subparsers.add_parser(
            command_name, help=command_description)
        command_obj.register(parser)



#!/usr/bin/env python3
"""This module tests the generate module"""
import getpass
import builtins
import unittest
import mock
import libpkpass.commands.cli as cli
import libpkpass.commands.generate as generate
from libpkpass.errors import CliArgumentError, DecryptionError
from .basetest.basetest import patch_args, ERROR_MSGS

class GenerateTests(unittest.TestCase):
    """This class tests the generate class"""
    def test_generate_success(self):
        """Test a successful generation of a password"""
        ret = True
        try:
            with patch_args(subparser_name='generate', identity='r3', nopassphrase='true',
                            pwname='gentest'):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    with mock.patch.object(getpass, 'getpass', lambda _: 'y'):
                        generate.Generate(cli.Cli())
        except DecryptionError:
            ret = False
        self.assertTrue(ret)

    def test_generate_no_pass(self):
        """test what happens with pwname is not supplied"""
        with self.assertRaises(CliArgumentError) as context:
            with patch_args(subparser_name='generate', identity='r3', nopassphrase='true',
                            pwname=None):
                with mock.patch.object(builtins, 'input', lambda _: 'y'):
                    with mock.patch.object(getpass, 'getpass', lambda _: 'y'):
                        generate.Generate(cli.Cli())
        self.assertEqual(context.exception.msg, ERROR_MSGS['pwname'])

if __name__ == '__main__':
    unittest.main()



"""This module allows for the population of external password stores"""
from os import path
# from yaml import safe_load, dump
from ruamel.yaml import YAML
from libpkpass.commands.show import Show
from libpkpass.crypto import puppet_password
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError

    ####################################################################
class Populate(Show):
    """This class implements the CLI functionality of password store integration"""
    ####################################################################
    name = 'populate'
    description = 'Populate external resource with password, currently supports: puppet_eyaml'
    selected_args = ['pwname', 'cabundle', 'card_slot', 'nopassphrase',
                     'certpath', 'color', 'identity', 'keypath',
                     'nocache', 'noverify', 'pwstore', 'quiet',
                     'theme_map', 'type', 'value', 'verbosity']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        # currently only support puppet
        pop_type = self.args['type']
        if pop_type not in ['puppet_eyaml']:
            raise CliArgumentError("'%s' is an unsupported population type" % pop_type)

        if self.args['pwname'] not in self.args['populate'][pop_type]['passwords'].keys():
            raise CliArgumentError("'%s' doesn't have a mapping" % pop_type)

        password = PasswordEntry()
        myidentity = self.identities.iddb[self.args['identity']]
        plaintext_pw = self._decrypt_wrapper(
            self.args['pwstore'],
            password,
            myidentity,
            self.args['pwname']
        ).strip()
        if pop_type == 'puppet_eyaml':
            self._handle_puppet(plaintext_pw)

        ####################################################################
    def _handle_puppet(self, plaintext_pw):
        ####################################################################
        directory = path.expanduser(self.args['populate']['puppet_eyaml']['directory'])
        pkcs7_pass = puppet_password(
            path.expanduser(self.args['populate']['puppet_eyaml']['bin']),
            plaintext_pw)
        if self.args['value']:
            print(pkcs7_pass)
        else:
            for hiera_file, names in self.args['populate']['puppet_eyaml']['passwords'][self.args['pwname']].items():
                with open(path.join(directory, hiera_file), 'r') as data_file:
                    print("Updating: %s" % hiera_file)
                    yaml = YAML()
                    yaml.indent(mapping=2, sequence=2, offset=2)
                    hiera_yaml = yaml.load(data_file.read())
                    # hiera_yaml = safe_load(data_file.read())
                    for name in names:
                        print("Updating: %s" % name)
                        hiera_yaml[name] = pkcs7_pass
                with open(path.join(directory, hiera_file), 'w') as data_file:
                    yaml.dump(hiera_yaml, data_file)
                    # data_file.write(YAML().dump(hiera_yaml))

        ####################################################################
    def _validate_args(self):
        ####################################################################
        for argument in ['pwname', 'keypath', 'type']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(
                    "'%s' is a required argument" % argument)

        ####################################################################
    def _decrypt_wrapper(self, directory, password, myidentity, pwname):
        """Decide whether to decrypt normally or for escrow"""
        ####################################################################
        password.read_password_data(path.join(directory, pwname))
        return self._decrypt_password_entry(password, myidentity)

        ####################################################################
    def _decrypt_password_entry(self, password, myidentity):
        """This decrypts a given password entry"""
        ####################################################################
        plaintext_pw = password.decrypt_entry(
            identity=myidentity, passphrase=self.passphrase, card_slot=self.args['card_slot'])
        if not self.args['noverify']:
            result = password.verify_entry(
                myidentity['uid'], self.identities)
            if not result['sigOK']:
                print("warning: could not verify that '%s' correctly signed your password entry." %
                      result['distributor'])
            if not result['certOK']:
                print("Warning: could not verify the certificate authenticity of user '%s'." %
                      result['distributor'])
        return plaintext_pw



"""This Module allows for editing metadata of passwords"""
from os import path
from libpkpass.commands.command import Command
from libpkpass.password import PasswordEntry
from libpkpass.errors import CliArgumentError

    ####################################################################
class Modify(Command):
    """This class implements the cli list"""
    ####################################################################
    name = 'modify'
    description = 'Modify the metadata of a password'
    selected_args = Command.selected_args + ['pwname', 'pwstore']

        ####################################################################
    def _run_command_execution(self):
        """ Run function for class.                                      """
        ####################################################################
        full_path = path.join(self.args['pwstore'], self.args['pwname'])
        password = PasswordEntry()
        password.read_password_data(full_path)
        editable = ['authorizer', 'description']
        for key, value in password['metadata'].items():
            if key in editable:
                print("%s %s" % (self.color_print("Current value for '%s':" % key, "first_level"), value))
                password['metadata'][key] = input("New Value for %s: " % key)

        password.write_password_data(full_path)

        ####################################################################
    def _validate_args(self):
        ####################################################################
        for argument in ['pwname', 'pwstore']:
            if argument not in self.args or self.args[argument] is None:
                raise CliArgumentError(
                    "'%s' is a required argument" % argument)
